import os
import random


def missing_msg(params, target):
    msg = ''
    if not params.issubset(target):
        msg += f'unexpected values in input: {params - target}; '
    if not target.issubset(params):
        msg += f'missing values in input: {target - params}'
    return msg


def random_id():
    """ Random ID identifier."""
    letters = list('ABDEFGHIJKLMNOPQRSTUVWXZYZ0123456789')
    id_ = [random.choice(letters) for _ in range(8)]
    return ''.join(id_)


def call_script(path, content, grab_output=False, cleanup=False):
    with open(path, 'w') as f:
        f.write(content)
    os.system(f'chmod +x {path}')
    if grab_output:
        output = os.popen(path).read()[:-1]
        if not output:
            raise Exception('grabbing output failed.')
    else:
        output = os.system(path)
    if cleanup:
        os.system(f'rm {path}')
    return output


def log_content(content):
    lines = content.split('\n')
    len_ = max([len(x) for x in lines])
    print(f'content:\n  ' + len_ * '-')
    print('\n'.join(['  ' + x for x in lines]))
    print(f'  ' + len_ * '-')



import glob
import json


def find_jd_files(depth=3):
    files = []
    for dep in range(depth):
        pattern = '*/' * dep + 'jd.json'
        files.extend(glob.glob(pattern))
    return files


def load_all_resources(root='', depth=3):
    """
    Load all of the meta data of all deployments.
    """
    all_files = find_jd_files(depth=depth)
    all_ = []
    for file_ in all_files:
        if not file_.startswith(root):
            continue
        with open(file_) as f:
            meta = json.load(f)
        all_.extend(meta)
    all_ = sorted(all_, key=lambda x: x['created'])
    return all_


def load_resource(id_):
    """
    Load meta data of id.

    :param id_: ID identified of deployment.
    """
    all_resources = load_all_resources()
    return next(x for x in all_resources if x['id'] == id_)



import datetime
import json
import os
import re

from jd.resources import load_resource, load_all_resources
from jd.utils import random_id
from jd.utils import missing_msg
from jd.templates import call_template, load_template, get_path


def get_project():
    return os.getcwd().split('/')[-1]


def prepare_params_for_resource(path, template, root, params):
    meta = {}
    meta['id'] = random_id()
    meta['commit'] = os.popen('git rev-parse HEAD').read().split('\n')[0]
    msg = os.popen('git log -1 --pretty=%B').read().split('\n')[0]
    meta['message'] = '\n'.join([x.strip() for x in msg.split('\n') if x.strip()])
    meta['jd_path'] = root + 'jd.json'
    if not meta['commit']:
        raise Exception('something went wrong determining the current commit')
    subdir = f'{root}.jd/{meta["id"]}'
    meta['subdir'] = subdir
    os.system(f'mkdir -p {meta["subdir"]}/tasks')
    meta['project'] = get_project()
    assert set(params.keys()) == set(template['params']), \
        missing_msg(set(params.keys()), set(template['params']))

    info = {'params': params,
            'config': template['config'],
            'created': str(datetime.datetime.now()),
            'template': path,
            **meta}

    try:
        with open(f'{meta["jd_path"]}') as f:
            all_jobs = json.load(f)
    except FileNotFoundError:
        all_jobs = []

    all_jobs.append(info)
    with open(f'{meta["jd_path"]}', 'w') as f:
        json.dump(all_jobs, f, indent=2)

    return info


def postprocess_params_for_resource(info):
    info['stopped'] = str(datetime.datetime.now())
    with open(info['jd_path']) as f:
        jobs = json.load(f)
    jobs = [x if x['id'] != info['id'] else info for x in jobs]
    with open(f'{info["jd_path"]}', 'w') as f:
        json.dump(jobs, f, indent=2)


def rm(id, purge=False, down=True):
    r = load_resource(id)
    if down:
        if 'stopped' not in r:
            build(r['template'], 'down', id=id)
    if purge:
        build(r['template'], 'purge', id=id)

    with open(r['jd_path']) as f:
        jobs = json.load(f)
    jobs = [x for x in jobs if x['id'] != r['id']]

    with open(r['jd_path'], 'w') as f:
        json.dump(jobs, f, indent=2)


def ls(template=None, root=''):
    out = load_all_resources(root=root)
    out = [{k: v for k, v in x.items() if k not in {'values', 'config'}} for x in out]
    if template is not None:
        out = [x for x in out if re.match(template, x['template']) is not None]
    print(json.dumps(out, indent=2))
    return out


def view(id):
    out = load_resource(id)
    print(json.dumps(out, indent=2))


def _get_last_id(template_path):
    records = ls(template=template_path)
    return records[-1]['id']


def _get_jd_path(id):
    records = ls()
    return next(x for x in records if x['id'] == id)['jd_path']


def build(path, method, id=None, root='', **params):
    """ Call template located at "path" with parameters.

    :param path: Template .yaml path.
    :param method: Name of build to run.
    :param root: Root directory for storing meta-data
    :param params: Run-time parameters (key values)
    """

    if id is None and not (method == 'up'):
        id = _get_last_id(path)
    if path is None:
        path = get_path(id=id)
    template = load_template(path)
    if root and root[-1] != '/':
        root = root.strip() + '/'

    try:
        if method == 'up':
            if not os.path.exists(root + '.jd'):
                os.makedirs(root + '.jd')
            info = prepare_params_for_resource(path, template, root, params)
        else:
            jd_path = _get_jd_path(id)
            assert id is not None
            with open(jd_path) as f:
                jobs = json.load(f)
            info = next(j for j in jobs if j['id'] == id)
            params = info['params']

        meta = {k: v for k, v in info.items() if k not in {'values', 'params', 'config'}}
        call_template(template, method, params, meta, on_up=method == 'up')

        if method == 'down':
            postprocess_params_for_resource(info)

    except Exception as e:
        # if method == 'up':
        #     rm(info['id'], down=False, purge=False)
        pass
        raise e




import json
from jinja2 import Template, StrictUndefined
from jd.utils import random_id, log_content, call_script


def get_or_create_values(template, params, meta, on_up=False):
    with open(meta['jd_path']) as f:
        jobs = json.load(f)
    info = next(j for j in jobs if j['id'] == meta['id'])
    existing_values = info.get('values', {})
    missing_values = {k: v for k, v in template.get('values', {}).items()
                      if k not in existing_values}

    if missing_values:
        existing_values.update(
            create_values(missing_values,
                          params,
                          meta,
                          template['config'],
                          existing_values=existing_values,
                          on_up=on_up)
        )
    info['values'] = existing_values
    with open(meta['jd_path'], 'w') as f:
        json.dump(jobs, f, indent=2)
    return existing_values


def create_values(values, params, meta, config, existing_values=None,
                  on_up=False):
    """
    Create values. Go through dictionary of values based on parameters dictionary.

    :param values: Dictionary of values with Jinja2 variables in strings.
    """
    if existing_values is None:
        existing_values = {}

    for k in values:
        if values[k]['type'] == 'static':
            if not on_up or values[k].get('on_up', True):
                existing_values[k] = \
                    create_static_value(values[k]['content'], existing_values, params, meta,
                                        config)
            else:
                print(f"didn't create static value because on_up=False: {k}")

    for k in values:
        if values[k]['type'].startswith('output/'):
            if not on_up or values[k].get('on_up', True):
                output = create_output_value(values[k]['content'], existing_values, params, meta,
                                             config)
                suffix = values[k]['type'].split('output/')[-1]
                if suffix == 'str':
                    existing_values[k] = output
                elif suffix == 'json':
                    existing_values[k] = json.loads(suffix)
                else:
                    raise NotImplementedError(f'output type for value not supported: {suffix}.')
            else:
                print(f"didn't create output value because on_up=False: {k}")
    return existing_values


def create_output_value(value, other_values, params, meta, config):
    script = Template(value, undefined=StrictUndefined).render(params=params,
                                                               meta=meta,
                                                               config=config,
                                                               values=other_values)
    id = random_id()
    print('creating value with script:')
    log_content(script)
    output = call_script(f'/tmp/{id}', script, grab_output=True, cleanup=True)
    return output


def create_static_value(value, other_values, params, meta, config):
    """
    Format a value using template parameters.

    :param value: Value to be formatted using Jinja2.
    :param other_values: Other pre-built values to be used in the value with {{ values[...] }}.
    :param params: Dictionary of parameters, referred to by {{ params[...] }}.
    """
    if isinstance(value, str):
        return Template(value, undefined=StrictUndefined).render(params=params,
                                                                 values=other_values,
                                                                 config=config,
                                                                 meta=meta)
    elif isinstance(value, list):
        return [create_static_value(x, other_values, params, config, meta) for x in value]

    elif isinstance(value, dict):
        return {k: create_static_value(value[k], other_values, params, meta, config)
                for k in value}

    else:
        raise NotImplementedError('only strings, and recursively lists and dicts supported')


import os

from jinja2 import Template, StrictUndefined
import yaml

from jd.values import get_or_create_values
from jd.resources import load_resource
from jd.utils import missing_msg, call_script, log_content


def get_path(id):
    return load_resource(id)['template']


def load_template(path):
    """
    Load deployment template.

    :params path: Load template from path. If templte has "parents" field, then combine with parent
        templates. Parameters named in binds are set in the parents as defaults.
    """
    if not path.endswith('.yaml'):
        path += '.yaml'
    with open('templates/' + path) as f:
        template = yaml.safe_load(f.read())

    assert 'builds' in template, 'template must have section "builds"'
    assert 'params' in template, 'template must have section "params"'
    assert 'config' in template, 'template must have section "config"'
    assert 'meta' in template, 'template must have section "meta"'
    return template


def call_template(template, method, params, meta, on_up=False):
    """
    Call template with parameters.

    :param template: Loaded template dictionary.
    :param method: Name of build to run.
    """

    assert set(params.keys()) == set(template['params']), \
        missing_msg(set(params.keys()), set(template['params']))

    def build_method(method):
        cf = template['builds'][method]
        values = get_or_create_values(template, params, meta, on_up=on_up)

        deploy_dir = f'{meta["subdir"]}/tasks'
        if not os.path.exists(deploy_dir):
            os.makedirs(deploy_dir, exist_ok=True)

        if cf['type'] != 'sequence':
            print(f'building "{method}"')

            content = Template(cf['content'], undefined=StrictUndefined).render(
                params=params, values=values, meta=meta, config=template['config'],
            )
            log_content(content)
            if cf['type'] == 'file':
                path = f'{deploy_dir}/{method}'
                with open(path, 'w') as f:
                    f.write(content)
            if cf['type'] == 'script':
                path = f'{deploy_dir}/{method}'
                exit_code = call_script(path, content, grab_output=False, cleanup=False)
                if exit_code and exit_code not in cf.get('whitelist', []):
                    raise Exception(f'script exited with non-zero exit code: {exit_code}.')
            return

        for m in cf['content']:
            build_method(m)

    build_method(method)


import setuptools

import pkg_resources
import pathlib
from distutils.util import convert_path

versions = {}
ver_path = convert_path('jd/version.py')
with open(ver_path) as ver_file:
    exec(ver_file.read(), versions)

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


def parse_requirements(filename):
    with pathlib.Path(filename).open() as requirements_txt:
        return [str(requirement)
                for requirement in pkg_resources.parse_requirements(requirements_txt)]


setuptools.setup(
    name="ai-jobdeploy",
    version=versions['__version__'],
    author="LF1",
    author_email="contact@lf1.io",
    description="AI job deployment templates with YAML and Jinja2.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/lf1-io/jobdeploy",
    packages=setuptools.find_packages(),
    setup_requires=[],
    license="Apache 2.0",
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
    ],
    python_requires='>=3.6',
    install_requires=parse_requirements('requirements.txt'),
    package_data={'': ['requirements.txt']},
    include_package_data=True,
    entry_points={
	'console_scripts': [
            'jd = jd.__main__:cli'
        ]
    }
)



import click
import re
from jd.controller import build as _build, rm as _rm, ls as _ls, view as _view


@click.group()
def cli():
    ...


def parse_inputs(x):
    groups = re.finditer('"([^\']+)"', x)
    reference = {}
    for i, g in enumerate(groups):
        x = x.replace(g.group(), f'#{i}')
        reference[f'#{i}'] = g.groups()[0]
    my_dict = dict([x.split('=') for x in x.split(',')])
    for k, val in my_dict.items():
        if val.isnumeric():
            my_dict[k] = eval(val)
        elif val.startswith('#'):
            my_dict[k] = reference[val]
        elif val in {'true', 'True', 'false', 'False'}:
            my_dict[k] = val.lower() == 'true'
        elif '+' in val:
            val = val.split('+')
            val = [x for x in val if x]
            val = [eval(x) if x.isnumeric() else x for x in val]
            my_dict[k] = val
    return my_dict


class KeyValuePairs(click.ParamType):
    """Convert to key value pairs"""
    name = "key-value-pairs"

    def convert(self, value, param, ctx):
        """
        Convert to key value pairs

        :param value: value
        :param param: parameter
        :param ctx: context
        """
        if not value.strip():
            return {}
        try:
            my_dict = parse_inputs(value)
            return my_dict
        except TypeError:
            self.fail(
                "expected string for key-value-pairs() conversion, got "
                f"{value!r} of type {type(value).__name__}",
                param,
                ctx,
            )
        except ValueError:
            self.fail(f"{value!r} is not a valid key-value-pair", param, ctx)


@cli.command()
@click.option('--template', default=None, help='type of resource to list')
@click.option('--root', default='', help='limit list to directory root')
def ls(template, root):
    _ls(template, root)


@cli.command()
@click.argument('id')
def view(id):
    _view(id)


@cli.command()
@click.argument('id')
@click.option('--purge/--no-purge', default=False, help='purge resource')
@click.option('--down/--no-down', default=True, help='tear down resource with the down method')
def rm(id, purge, down):
    _rm(id, purge, down)


@cli.command(help='build template')
@click.argument('method')
@click.option('--template', default=None)
@click.option('--id', default=None)
@click.option('--params', default=None, help='key-value pairs to add to build',
              type=KeyValuePairs())
@click.option('--root', default='')
def build(method, template, id, params, root):
    print(params)
    if params is None:
        params = {}
    if isinstance(template, str) and template.endswith('.yaml'):
        template = template.split('.yaml')[0]
    _build(template, method, id=id, root=root, **params)


if __name__ == '__main__':
    cli()